#! /bin/env python

import wave, sys, struct, numpy, math, pprint

def usage( exit=True ):
    print """
usage : python wav_compare.py file1.wav file2.wav
"""
    if exit:
        sys.exit(-1)

class Wave( object ):
    def __init__( self, fname ):
        self.fname = fname
        self.wave = wave.open( fname, "r" )
        self.skipped_blanks = None
    
    def print_info( self ):
        nchannels, sampwidth, framerate, nframes, comptype, compname = self.wave.getparams()
        print """
File name : <%s>
    Channels         : %d
    Bytes per sample : %d
    Sample rate      : %d Hz
    Length           : %.05f s [%d samples]
    Encoding         : %s
    
""" % (self.fname, nchannels, sampwidth, framerate, nframes*1.0/framerate, nframes, compname)

    def get_size( self ):
        return self.wave.getnframes() * self.wave.getsampwidth()

    def skip_silence( self ):
        zero = "\x00" * (self.wave.getnchannels() * self.wave.getsampwidth())
        self.wave.rewind()
        if self.skipped_blanks is None:
            while self.wave.readframes(1) == zero:
                pass
            self.skipped_blanks = self.wave.tell() - 1
            print "Skipped %d silent samples at start of <%s>" % (self.skipped_blanks, self.fname)
        else:
            self.wave.setpos( self.skipped_blanks )
        return self.skipped_blanks
    
    def read_block( self, n ):
        r = numpy.fromstring( self.wave.readframes( n ), "<u2" ).astype('d')
        return r.reshape( len(r)/self.wave.getnchannels(), self.wave.getnchannels() )
    
    def is_silence_until_end( self ):
        while True:
            d = self.wave.readframes(4096).strip("\x00")
            if d:
                return False
        return True
            

def run():
    print
    try:
        fname_a, fname_b = sys.argv[1:]
    except:
        usage()

    try:
        wa = Wave( fname_a )
        wb = Wave( fname_b )
    except Exception, e:
        print e
        usage()
    
    wa.print_info()
    wb.print_info()
    
    # Check compatibility and display some useful crap
    if wa.wave.getframerate() != wa.wave.getframerate():
        print "Sample rates are different, cannot compare."
        return
    
    if wa.wave.getsampwidth() != wb.wave.getsampwidth():
        print "Sample width is different, cannot compare."
        return

    if wa.wave.getsampwidth() != 2 or wb.wave.getsampwidth() != 2:
        print "This only works with 16-bit WAVs."
        return

    if wa.wave.getnchannels() != wa.wave.getnchannels():
        print "Warning : Number of channels is different. I will compare each channel of file A with each channel of file B."

    for channel_a in xrange( wa.wave.getnchannels() ):
        for channel_b in xrange( wb.wave.getnchannels() ):
            print "Comparing channel %d from <%s> and channel %d from <%s>" % (channel_a, wa.fname, channel_b, wb.fname )

            # Do the check
            wa.skip_silence()
            wb.skip_silence()

            # Try to cross-correlate to find the offsets
            print "Looking for offsets..."
            wa.read_block( 16384 )
            wb.read_block( 16384 )
            da = wa.read_block( 8192 )
            db = wb.read_block( 8192 )
    
            if len( da ) != len( db ) or len(da) < 4096 or len(db) < 4096:
                print "Come on, these files are too short, I need at least 32k samples..."
                return
        
            dists = {}
            l = 4096
            #~ da = da - numpy.mean(da)
            #~ db = db - numpy.mean(db)
            #~ da = da / math.sqrt( numpy.sum( da**2 ))
            #~ db = db / math.sqrt( numpy.sum( db**2 ))
            
            for offset in xrange( -4000,4000 ):
                if offset < 0:
                    pa = -offset
                    pb = 0
                else:
                    pa = 0
                    pb = offset
                a = da[pa:(pa+l),channel_a]
                b = db[pb:(pb+l),channel_b]
                dists[offset] = math.sqrt(numpy.sum( (a - b)**2 ))
                

            print "Found the following candidate offsets"
            dists = sorted( dists.items(), key=lambda x: (x[1],abs(x[0])) )
            for offset, dist in dists[:10]:
                print "offset %6d, distance" % (offset), dist
            
            used_offset = dists[0][0]
            used_dist = dists[0][1]
            print "Will use offset : %d samples" % used_offset
            n = sum( 1 for offset, dist in dists if dist == used_dist )
            if n > 10:
                print "I found %d possible offsets, is this a sinewave or something ? I will use the smallest offset." % n

            # rewind files
            wa.skip_silence()
            wb.skip_silence()
            
            # skip offset

            if used_offset > 0:
                wb.read_block( used_offset )
                print "Skipping %d samples at start of <%s>" % (used_offset, wb.fname)
            elif used_offset < 0:
                wa.read_block( -used_offset )
                print "Skipping %d samples at start of <%s>" % (-used_offset, wa.fname)

            errors = 0
            samples = 0
            ne = 0

            # Try to cross-correlate to find the offsets
            while True:
                da = wa.read_block( 32 )
                db = wb.read_block( 32 )
                
                if not( len( da ) and len( db )):
                    break
                    
                m = min( len( da ), len( db ))
                samples += m
                e = numpy.sum( db[:m,channel_b] != da[:m,channel_a] )
                errors += e
                if e:
                    ne += 1
                    if ne < 10:
                        print "Found some errors."
                        print da[:m,channel_a]
                        print db[:m,channel_b]
                    
                
                if len( da ) != len( db ):
                    break
            
            if len( da ) != len( db ):
                print "Comparison ended prematurely because one of the files is shorter than the other :\n"
                for d,w in (da,wa),(db,wb):
                    r = w.wave.getnframes() - w.wave.tell() + len(d) - m
                    print "%s unexamined samples remaining in <%s>" % (r, w.fname)
                
            
            print "** Result **"
            print "*** Samples compared : %d" % samples
            print "*** Channel %d of <%s> versus Channel %d of <%s> :" % (channel_a, wa.fname, channel_b, wb.fname)
            print "*** %d differences" % errors
            print
            print
            





run()
















